// Copyright © 2016 Université Paris-Sud, Written by Lénaïc Bagnères, lenaic.bagneres@u-psud.fr

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


#ifndef EXPRESSION_HPP
#define EXPRESSION_HPP

#include <iostream>
#include <string>
#include <memory>
#include <sstream>
#include <algorithm>
#include <cctype>

#include "hopp/conversion/is_integer.hpp"


// Expression

class expression_t_
{
public:
	
	virtual ~expression_t_() = default;
	
	virtual void display(std::ostream & out = std::cout) const = 0;
	
	virtual double eval() const = 0;
};


// Expression (do not care about virtual & co)

class expression_t : public expression_t_
{
public:
	
	std::unique_ptr<expression_t_> expression;
	
public:
	
	expression_t() = default;
	
	template <class T> // T is an expression_t_
	expression_t(std::unique_ptr<T> && p) : expression(std::move(p)) { }
	
	template <class T> // T is an expression_t_
	expression_t(T && t) : expression(std::make_unique<T>(std::move(t))) { }
	
	inline
	expression_t(std::string const & expression);
	
	std::unique_ptr<expression_t_> release()
	{
		return std::unique_ptr<expression_t_>(expression.release());
	}
	
	bool is_null() const { return expression == nullptr; }
	
	virtual void display(std::ostream & out = std::cout) const
	{
		if (expression) { expression->display(out); }
	}
	
	virtual double eval() const
	{
		if (expression) { return expression->eval(); }
		else { return 0.0; } // we can throw
	}
};


// Constant

class constant_t : public expression_t_
{
public:
	
	int value;
	
public:
	
	constant_t(int const value) : value(value) { }
	
	virtual void display(std::ostream & out = std::cout) const
	{
		// TODO
	}
	
	virtual double eval() const
	{
		return 0.0; // TODO
	}
};


// Variable

class variable_t : public expression_t_
{
public:
	
	std::string name;
	
	int value;
	
public:
	
	variable_t(std::string const & name, int const value) : name(name), value(value) { }
	
	virtual void display(std::ostream & out = std::cout) const
	{
		// TODO
	}
	
	virtual double eval() const
	{
		return 0.0; // TODO
	}
};


// Operator

class operator_t : public expression_t_
{
public:
	
	char symbol; // Operator
	
	expression_t a; // left
	
	expression_t b; // right
	
public:
	
	operator_t(char const symbol, expression_t && a, expression_t && b) :
		symbol(symbol), a(std::move(a)), b(std::move(b)) { }
	
	operator_t(operator_t &&) = default; // TODO
	
	virtual void display(std::ostream & out = std::cout) const
	{
		// TODO
	}
	
	virtual double eval() const
	{
		return 0.0; // TODO
	}
};


// Definition

expression_t::expression_t(std::string const & expression)
{
	std::istringstream ss(expression);
	
	std::string word;
	
	while (ss >> word)
	{
		std::cout << "'" << word << "'" << std::endl;
		
		// TODO
	}
}

#endif
